/*
K1n9_Duk3's IMF crusher - Removes dummy commands from IMF files.
Copyright (C) 2013 K1n9_Duk3

Based on K1n9_Duk3's MIDI to IMF converter (mid2imf.exe)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned short UINT16;
typedef unsigned int UINT32;
typedef unsigned char UINT8;

UINT32 IMF_RemoveDummyCommands(UINT8 *data, UINT32 *size)
{
	static UINT8 regs[256], reg_defined[256];
	UINT32 newsize = 0, pos=0, i;
	UINT32 oldsize = *size;

	memset(regs, 0, 256);
	memset(reg_defined, 0, 256);
	while(pos < *size)
	{
		UINT16 newpause;
		UINT8 reg = data[pos++];
		if (((reg >= 0xB0 && reg <= 0xB8) || reg == 0xBD) && reg_defined[reg] && newsize > 0)
		{
			// don't remove NoteOn/Off commands:
			data[newsize++] = reg;
			data[newsize++] = regs[reg];
			data[newsize++] = 0;
			data[newsize++] = 0;
		}
		reg_defined[reg] = 1;
		regs[reg] = data[pos++];
		newpause = data[pos] | (data[pos+1] << 8);
		pos += 2;
		if(newpause || (pos == *size))
		{
			// write data for all registers:
			for(i=0; i<256; i++)
			{
				if(!reg_defined[i])
					continue;

				data[newsize++] = i;
				data[newsize++] = regs[i];
				data[newsize++] = 0;
				data[newsize++] = 0;

				reg_defined[i] = 0;
			}
			data[newsize-2] = newpause & 0xFF;
			data[newsize-1] = (newpause >> 8) & 0xFF;
		}
	}
	*size = newsize;
	return oldsize-newsize;
}

UINT32 IMF_RemoveRedundantCommands(UINT8 *data, UINT32 *size)
{
	static UINT8 regs[256], reg_defined[256];
	UINT16 newpause, lastpause=0;
	UINT32 newsize = 0, pos=0;
	UINT32 oldsize = *size;

	memset(regs, 0, 256);
	memset(reg_defined, 0, 256);

	while (pos < *size)
	{
		newpause = data[pos+2] | (data[pos+3] << 8);
		if(reg_defined[data[pos]] && (regs[data[pos]] == data[pos+1]))
		{
			newpause += lastpause;
			data[newsize-2] = newpause & 0xFF;
			data[newsize-1] = (newpause >> 8) & 0xFF;
			pos += 4;
		} else {
			reg_defined[data[pos]] = 1;
			regs[data[pos]] = data[pos+1];
			// copy imf data:
			data[newsize++] = data[pos++];
			data[newsize++] = data[pos++];
			data[newsize++] = data[pos++];
			data[newsize++] = data[pos++];
		}
		lastpause = newpause;
	}
	*size = newsize;
	return oldsize-newsize;
}

void IMF_ConvertRate(UINT8 *data, UINT32 size, UINT16 fromRate, UINT16 toRate)
{
	double fracPause = 0.0;
	UINT16 pause;

	data++;
	data++;
	while (size)
	{
		pause = data[0] | (data[1] << 8);
		fracPause += (double)(pause*toRate)/fromRate;
		pause = UINT16(fracPause);
		fracPause -= pause;
		data[0] = pause & 0xFF;
		data[1] = (pause >> 8) & 0xFF;
		data += 4;
		size -= 4;
	}

}

int IMF_WriteIMF(char* filename, UINT8 *data, UINT32 size, UINT8 chunk)
{
	FILE *out = fopen(filename, "wb");
	if (out)
	{
		if(chunk)
		{
			UINT8 buf[2];
			buf[0] = size & 0xFF;
			buf[1] = (size >> 8) & 0xFF;
			fwrite(buf, 1, 2, out);
		}
		fwrite(data, 1, size, out);
		fclose(out);
		return 0;
	} else {
		printf("ERROR: could not create %s\n", filename);
		return -1;
	}
}

UINT16 ReadInt16LE(FILE* in)
{
	UINT8 buf[2];
	fread(buf, 1, 2, in);
	return (buf[0] | (buf[1] << 8));
}

int IMF_IsChunk(FILE* in)
{
	UINT16 chunksize, buff, i=42;
	UINT32 sum1=0, sum2=0;

	if(!in)
		return 0;

	if (!feof(in))
	{
		chunksize = ReadInt16LE(in);
		if ((chunksize == 0) || (chunksize & 3))
			return 0;

		while(!feof(in) && i)
		{
			fread(&buff, sizeof(buff), 1, in);
			sum1 += buff;
			fread(&buff, sizeof(buff), 1, in);
			sum2 += buff;
			i--;
		}
		fseek(in, 0, SEEK_SET);
		return (sum1 > sum2);
	}
	return 0;
}

static UINT8 *imfdata = NULL;
static UINT32 imfsize = 0;

int IMF_LoadIMF(char* filename)
{
	UINT32 size;
	FILE *in = fopen(filename, "rb");
	if (in)
	{
		if (IMF_IsChunk(in))
		{
			size = ReadInt16LE(in);
		} else {
			fseek(in, 0, SEEK_END);
			size = ftell(in);
			size -= size & 3;	// only multiples of 4 bytes allowed!
			fseek(in, 0, SEEK_SET);
		}
		if(imfdata)
			free(imfdata);
		imfdata = (UINT8*) malloc(size);
		if(imfdata)
		{
			fread(imfdata, 1, size, in);
			imfsize = size;
		} else {
			printf("ERROR: out of memory while reading %s\n", filename);
			fclose(in);
			return -1;
		}
		fclose(in);
		return 0;
	} else {
		printf("ERROR: could not open %s\n", filename);
		return -1;
	}
}

void main(int argc, char* argv[])
{
	printf("K1n9_Duk3's IMF crusher v1.0\n");
	printf("============================\n\n");

	if(argc < 3)
    {
        printf("Usage: IMFCRUSH <infile> <outfile> [OPTIONS]\n\n");
		printf("Recognized options are:\n"
			   " /keepdummy     keep instructions that get overwritten in the same cycle\n"
			   " /keepredundant keep instructions that write the same value that's already in\n"
               "                the AdLib register\n"
			   " /keepall       combination of /keepdummy and /keepredundant\n"
			   " /chunk         saves an AUDIOT chunk instead of a plain IMF file\n"
			   " \n"
			   " /convert <fromRate> <toRate>\n"
			   "                converts the song to a different rate\n"
			   " \n"
			   );

	} else {
		int convert=0, fromRate, toRate;
		int keepd=0, keepr=0, chunk=0;

		#define IFARG(str) if(!_stricmp(arg, (str)))

	    for(int i = 3; i < argc; i++)
		{
			char* arg = argv[i];
			IFARG("/keepdummy")
				keepd=1;
			else IFARG("/keepredundant")
				keepr=1;
			else IFARG("/keepall")
				keepd=keepr=1;
			else IFARG("/chunk")
				chunk = 1;
			else IFARG("/nochunk")
				chunk = 0;
			else IFARG("/convert")
				if (i+2 >= argc)
				{
					printf("ERROR: /convert needs 2 arguments\n");
					return;
				} else {
					convert = 1;
					fromRate = atoi(argv[++i]);
					if (fromRate <= 0)
					{
						printf("ERROR: <fromRate> must be greater than 0\n");
						return;
					}
					toRate = atoi(argv[++i]);
					if (toRate <= 0)
					{
						printf("ERROR: <toRate> must be greater than 0\n");
						return;
					}
				}
			else
			{
				printf("ERROR: unhandled option: %s\n", arg);
				return;
			}
		}

		printf("Loading IMF song from %s\n", argv[1]);
		if (IMF_LoadIMF(argv[1]) == 0)
		{
			if (convert)
			{
				printf("Converting IMF from %d Hz to %d Hz\n", fromRate, toRate);
				IMF_ConvertRate(imfdata, imfsize, fromRate, toRate);
			}
			if (!keepd)
			{
				UINT32 tmp = IMF_RemoveDummyCommands(imfdata, &imfsize);
				printf("RemoveDummyCommands removed %u bytes (%u commands).\n", tmp, tmp/4);
			}
			if(!keepr)
			{
				UINT32 tmp = IMF_RemoveRedundantCommands(imfdata, &imfsize);
				printf("RemoveRedundantCommands removed %u bytes (%u commands).\n", tmp, tmp/4);
			}

			printf("Writing IMF song to %s", argv[2]);
			if(chunk)
				printf(" (as AUDIOT chunk)\n");
			else
				printf(" (as plain IMF data)\n");

			IMF_WriteIMF(argv[2], imfdata, imfsize, chunk);
		}
	}
}